/***********************************************************************************************
 * BikeLight.c - A LumiLED bicycle light controlled by an AVR Tiny13
 *
 * This program uses an AVR Tiny13 to control a LumiLED 3021-X-[E|I]-XXXX BuckPuck LED
 * power module with power supplied via a external rechageable battery.
 *
 * The 3021 BuckPuck with Control/Dimming capability can adjust the output current
 * to the LED based on a 0-5V control signal.  Based on Figure 2 in the 3021 datasheet,
 * full output current is available when ctrl <= 1.65V +/-5% and output current is 0 when
 * ctrl is >= 4.2V +/-5%.  LED current is linear between 1.65V and 4.2V.  However, this does
 * not seem to be the case when working with a PWM signal.  A 0-5V PWM signal provides full
 * control from 0% to a 100% duty cycle.
 *
 * It should be noted that the frequency of the PWM signal does not appear make a difference
 * in the efficiency of the BuckPuck; it DOES make a difference in the response curves.
 * A higher frequency PWM signal flattens the output curve at the low end giving finer control
 * over the output current.  See the Excel spreadsheet "LuxDrive BuckPuck Current.xls" for
 * additional data.
 *
 * A single pushbutton is used to control a software implemented state machine.  Currently
 * there are three steady brightness levels and two pulsed brightness levels.  The pulsed
 * modes are used to increase other's awareness of the rider without completely extinguishing
 * illumination as some flashing bike lights do.  It is hoped that this will still increase
 * the riders visibiility to others without the rider having to deal with his path
 * illumination periodically going completely off.
 *
 * The BikeLight remembers the last used mode before power off and returns to that mode
 * after the first button press after power on.  A shortcut mode enables one to quickly
 * go to steady full on mode at any time without cycling through the remaining modes.
 *
 * The shortcut mode toggles between steady full on and the previously selected mode.  To
 * enter the shortcut mode, the pushbutton is held for SHORT_HOLD milliseconds.  One
 * SHORT_HOLD press turns on steady full on mode, and the next SHORT_HOLD press reverts
 * to the previous mode. 
 *
 * To turn off the LED, press and hold the button LONG_HOLD milliseconds.  The BikeLight is
 * in standby mode in this state and will continue to draw approximately 7ma  To completely
 * power down the BikeLight, disconnect the battery.
 *
 *
 * Programmer:  Tom Blough <tom(at)thebloughs(dot)net>
 * Version:     Original 2008/09/14
 * Website:     http://thebloughs.net/hobbies/electronics/bikeLight/
 *
 *
 * Copyright (c) 2008 Tom Blough
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 **********************************************************************************************/

#include <util/delay.h>
#include <avr/eeprom.h>
#include "utility.h"

// ATtiny13 running on internal oscillator 9.6Mhz/Clkdiv8/no additional startup time
// be sure to set project frequency to 1.2MHz for proper timer operation


// Physical connections
#define MODE_BTN(reg) BIT(B,1,reg)  // mode pushbutton
#define PWM_PIN(reg) BIT(B,0,reg)  // OCR0A pin on Tiny13 connected to buckpuck ctrl
#define PWM_OUTPUT OCR0A

// 8-bit PWM duty cycles for different BuckPuck output levels
#define	PWM_HIGH 0    // LED: 670mA   Iin: 508mA@18V ~ 3 hours with a 1500mAH battery
#define PWM_MED  190  // LED: 320mA   Iin: 228mA@18V ~ 6 hours         "
#define PWM_LOW  247  // LED:  40mA   Iin:  32mA@18V ~ 46 hours        "
#define PWM_OFF  255  // LED:   0mA   Iin:   7mA@18V ~ 214 hours       "

// State machine
#define MODE_HIGH 0
#define MODE_MED 1
#define MODE_LOW 2
#define MODE_MED_P 3  // medium with high pulse
#define MODE_LOW_P 4  // low with high pulse

#define NUM_MODES 5  // total number of possible states excluding "off"
#define MODE_OFF 255

// Timing Control
#define SCAN_TIME 50  // number of ms between scan cycles - max 255
#define PULSE_HIGH_TIME 1  // pulsing high time in scan cycles - max 255
#define PULSE_LOW_TIME 20  // pulsing low time in scan cycles - max 255
#define LONG_HOLD 40  // scan cycles for "long" button hold period - max 255
#define SHORT_HOLD 20  // scan cycles for "short" button hold period - max 255

// Persistant Storage
uint8_t EEMEM nvMode;  // Non volatile storage for last known mode

// Global Variables
uint8_t gPWMHi = PWM_OFF;  // active pwm values
uint8_t gPWMLo = PWM_OFF;
uint8_t gPulsePhase = LOW;  // high or low phase of pulse cycle
uint8_t gPulseCount = 0;  // remaining cycles in current pulse phase


// This is our state machine
uint8_t setState( uint8_t state, uint8_t save_change)
{
	if( save_change && (state != MODE_OFF))  // save the new mode for next turn on?
	{
		eeprom_write_byte( &nvMode, state);	// store the new mode in persistant storage
	}

	// state machine
	switch( state)
	{
		case MODE_HIGH:
		{
			gPWMHi = PWM_HIGH;
			gPWMLo = PWM_HIGH;
			break;
		}
		case MODE_MED:
		{
			gPWMHi = PWM_MED;
			gPWMLo = PWM_MED;
			break;
		}
		case MODE_LOW:
		{
			gPWMHi = PWM_LOW;
			gPWMLo = PWM_LOW;
			break;
		}
		case MODE_MED_P:
		{
			gPWMHi = PWM_HIGH;
			gPWMLo = PWM_MED;
			break;
		}
		case MODE_LOW_P:
		{
			gPWMHi = PWM_MED;
			gPWMLo = PWM_LOW;
			break;
		}
		case MODE_OFF:
		default:
		{
			gPWMHi = PWM_OFF;
			gPWMLo = PWM_OFF;
		}
	}
	gPulseCount = 0;
	gPulsePhase = LOW;

	return state;
}


int main( void)
{
	uint8_t buttonCount = 0;  // number of cycles button has been held
	uint8_t shortcutPrev = MODE_HIGH;  // previous mode for Shortcut
	uint8_t shortcutOn = FALSE;  // flag for shortcut change
	uint8_t state;  // current mode

	// configure the push button
	MODE_BTN( DDR) = IN;
	MODE_BTN( PORT) = ON; //pullup on

	// set up PWM for intensity control
	PWM_PIN( DDR) = OUT;  // pin as output
	TCCR0A = (1 << WGM00) | (1 << WGM01) | (1 << COM0A1);  // Fast PWM, CTC+TOP
	TCCR0B = (1 << CS00);  // no prescale
	PWM_OUTPUT = PWM_OFF;

	// set the startup mode
	state = setState( MODE_OFF, FALSE);

	// main program loop - executed every SCAN_TIME ms
	while( TRUE)
	{
		// check button status
		if( MODE_BTN( PIN) == LOW) // button pushed
		{
			buttonCount++;

			if( buttonCount > LONG_HOLD)  // holding to turn off?
			{
				// clear the shortcut
				shortcutPrev = MODE_HIGH;

				// change mode and update state indicator
				state = setState( MODE_OFF, FALSE);
			}
			else if( (buttonCount > SHORT_HOLD) && (! shortcutOn))  // high brightness shortcut?
			{
				if( shortcutPrev == MODE_HIGH)  // waiting to go full on?
				{
					// change mode, but don't update state indicator
					setState( MODE_HIGH, FALSE);
					shortcutPrev = state;
					shortcutOn = TRUE;
				}
				else
				{
					// restore the previous state
					setState( shortcutPrev, FALSE);
					shortcutPrev = MODE_HIGH;
					shortcutOn = TRUE;
				}
			}
		}
		else  // button released, check if state change and process
		{
			if( buttonCount > SHORT_HOLD)  // button released after one of the long pushes?
			{
				shortcutOn = FALSE;
			}
			else if( buttonCount > 0)  // regular button press so state change required
			{
				// is this first button push after being off?
				if( state == MODE_OFF)  // restore previous mode
				{
					// if NVRAM is corrupted, we could get strange numbers here, but since we
					// mod the value, it will still result in a valid mode, albiet a random one.
					// We'll live with it once in a while, as long as it is not "off".
					state = eeprom_read_byte( &nvMode) % NUM_MODES;
					if( state == MODE_OFF) state = 0;
				}
				else  // change to next state
				{
					state = (state + 1) % NUM_MODES;
				}

				setState( state, TRUE);
			}
			buttonCount = 0;
		}

		// finally, set the new PWM value to control the LED brightness
		if( gPulseCount > 0)  // time remaining in current phase?
		{
			gPulseCount--;
		}
		else  // change pulse phase
		{
			if( gPulsePhase == LOW)
			{
				gPulseCount = PULSE_HIGH_TIME;
				gPulsePhase = HIGH;
				PWM_OUTPUT = gPWMHi;
			}
			else
			{
				gPulseCount = PULSE_LOW_TIME;
				gPulsePhase = LOW;
				PWM_OUTPUT = gPWMLo;
			}
		}

		// sleep until the next scan
		_delay_ms( SCAN_TIME);
	}
}
